#!/usr/bin/env python

"""Tests for `datasets` package."""


import unittest

from datasets.utils.token_dicts import TokenDicts
import tensorflow as tf


class TestDatasets(unittest.TestCase):
    """Tests for `datasets` package."""

    def setUp(self):
        if tf.__version__.startswith('1.'):
            tf.enable_eager_execution()
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_raw_query_float_dataset(self):
        """Test something."""
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        assert token_dicts is not None


if __name__ == '__main__':
    unittest.main()



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""raw multi-inputs dataset generator class
"""
from __future__ import absolute_import, division, print_function
import numpy as np

from datasets.utils import data_processor_dicts
from datasets.utils.common_struct import to_array_pad
from datasets.textline_parser import TextlineParser


class FastSeqParser(TextlineParser):
    """ parser
    """

    def __init__(self, token_dicts, feature_field_list, label_field, additive_schema=['id', 'sid'], weight_fn=None, max_sessions=5):
        assert additive_schema[0] == 'id', 'additive_schema[0] must be id'
        assert additive_schema[1] == 'sid', 'additive_schema[1] must be sid'
        self.max_sessions = max_sessions
        super().__init__(token_dicts, feature_field_list, label_field, additive_schema, weight_fn)
        self.feature_field_list = [x for x in feature_field_list if x.processor is not None]
        self.session_features, self.lines = self.init_session()

    def init_session(self):
        session = [[] if field.has_session else None for field in self.feature_field_list]
        return session, []

    def make_session(self):
        for i in range(len(self.lines)):
            if self.lines[i] is None:
                continue
            _, _, features, _ = super().parse(self.lines[i])
            for s, x in zip(self.session_features, features):
                if s is not None:
                    s[i] = x
            self.lines[i] = None

    def update_session(self, features, line):
        for s in self.session_features:
            if s and len(s) == self.max_sessions:
                s.pop(0)
        if len(self.lines) == self.max_sessions:
            self.lines.pop(0)
        if features is None:
            for s in self.session_features:
                if s is not None:
                    s.append(None)
            self.lines.append(line)
        else:
            for s, x in zip(self.session_features, features):
                if s is not None:
                    s.append(x)
            self.lines.append(None)

    def parse(self, line):
        last_sid = None
        weight = self.weight_fn(line)
        result = None
        features = None
        if weight > 0:
            label, additive_info, features, weight = super().parse(line)
            sid = additive_info[1]
            if last_sid is not None and sid != last_sid:
                self.session_features = self.init_session()
            last_sid = sid
            self.make_session()
            session_features = [to_array_pad(s + [f]) if s is not None else f for s, f in zip(self.session_features, features)]
            result = label, additive_info, session_features, weight
        self.update_session(features, line)
        return result



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""raw multi-inputs dataset generator class
"""
from __future__ import absolute_import, division, print_function
from datasets.utils.common_struct import get_variable_shape_index, is_batch_padding


class Parser(object):
    """ parser
    """

    def __init__(self, token_dicts, feature_field, label_field=None, additive_schema=['id'], weight_fn=None):
        self.token_dicts = token_dicts
        self.weight_fn = weight_fn
        assert feature_field is not None, 'feature_field must not none'
        if not isinstance(feature_field, list):
            feature_field = [feature_field]
        self.feature_field = feature_field
        if label_field is None:
            label_field = []
        elif not isinstance(label_field, list):
            label_field = [label_field]
        self.label_field = label_field
        self.additive_schema = additive_schema
        self.has_label = len(label_field) > 0
        self.data_field_list = []
        self.data_field_list.extend(label_field)
        self.data_field_list .extend(feature_field)
        # 储存边长shape的下标列表 e.g. [(12,None), (23,None,None)] 则 var_shape_index_list = [(1),(1,2)]
        self.var_shape_index_list = [get_variable_shape_index(
            x.shape) if x.is_with_len else None for x in self.data_field_list if x.processor is not None]
        self.label_range = 0
        if self.has_label:
            # print(label_field)
            for field in label_field:
                if field.processor is not None:
                    if field.is_with_len:
                        self.label_range += 2
                    else:
                        self.label_range += 1
                # print('range',self.label_range, field)
        self.flat_feature_names = []
        self.features_dict = {}
        for field in feature_field:
            if field.processor is not None:
                assert field.name != 'weight', 'weight is used, please change another field name'
                self.features_dict[field.name] = field
                self.flat_feature_names.append(field.name)
                if field.is_with_len:
                    self.flat_feature_names.append(field.name + '_len')
        self.label_dict = {}
        for field in label_field:
            if field.processor is not None:
                assert field.name != 'weight', 'weight is used, please change another field name'
                self.label_dict[field.name] = field
                self.flat_feature_names.append(field.name)
                if field.is_with_len:
                    self.flat_feature_names.append(field.name + '_len')
        if weight_fn is not None:
            self.flat_feature_names.append('weight')
        self.is_data_padding = self.is_padding(self.data_field_list)

    def set_weight_fn(self, weight_fn):
        self.weight_fn = weight_fn

    def is_padding(self, field_list):
        for field in field_list:
            if is_batch_padding(field.shape):
                return True
        return False

    def parse(self, line):
        raise NotImplementedError



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from collections import namedtuple
import subprocess
import os
import tensorflow as tf
import numpy as np

__all__ = ['DataSchema', 'load_local_filelist', 'type2tf_dict', 'load_hdfs_filelist', 'data_schemas2types',
           'data_schemas2shapes', 'is_batch_padding', 'get_variable_shape_index']

# shape : [12,10,12]
param = ["name", "processor", "dtype", "shape", "token_dict_name", "is_with_len", "max_len", "has_session"]
DataSchema = namedtuple('DataSchema', field_names=param)
DataSchema.__new__.__defaults__ = tuple([None] * len(param))

type2tf_dict = {
    "int": tf.int32,
    "int32": tf.int32,
    "int64": tf.int64,
    "float": tf.float32,
    "float32": tf.float32,
    "float64": tf.float64
}


def to_array_pad(array, dtype='int32'):
    """
    padding 多维动态数据(padding zeros)
    """
    len_array = [len(d) for d in array]
    pad_len = max(len_array)
    return np.asarray([np.concatenate((d, np.zeros((pad_len - l,), dtype=dtype)), axis=0) if pad_len > l else d for d, l in zip(array, len_array)])

    """
def to_array_pad_org(array):
    padding 多维动态数据(padding zeros)
    len_array = [len(d) for d in array]
    pad_len = max(len_array)
    return np.asarray([np.pad(d, (0, pad_len - l), 'constant') if pad_len > l else d for d, l in zip(array, len_array)])
    """


def load_hdfs_filelist(datadir, file_suffix, hadoop):
    filenames = []
    cmd = hadoop + ' fs -ls ' + datadir
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True) as pipe:
        for line in iter(pipe.stdout.readline, b''):
            items = line.decode('utf8', 'ignore').strip('\n').split(' ')
            if len(items) > 3:
                if file_suffix is None or items[-1].split('/')[-1].endswith(file_suffix):
                    filenames.append(items[-1])
    return filenames


def load_local_filelist(datadir, file_suffix):
    filenames = os.listdir(datadir)
    filenames = map(lambda x: os.path.join(datadir, x), filenames)
    if file_suffix is None:
        filenames = [f for f in list(filenames)]
    else:
        filenames = [f for f in list(
            filenames) if f.endswith(file_suffix)]
    return filenames


def data_schemas2types(datafield_list):
    def data_schema2types(data_schema):
        types = type2tf_dict[data_schema.dtype]
        if data_schema.is_with_len:
            return [types, tf.int32]
        else:
            return types

    if isinstance(datafield_list, DataSchema):
        return data_schema2types(datafield_list)
    elif isinstance(datafield_list, list):
        types = []
        for field in datafield_list:
            if field.processor is None:
                continue
            type = data_schema2types(field)
            if isinstance(type, list):
                types.extend(type)
            else:
                types.append(type)
        return types


def data_schemas2shapes(datafield_list):
    def data_schema2shapes(data_schema):
        shapes = data_schema.shape
        if data_schema.is_with_len:
            var_list = get_variable_shape_index(field.shape)
            if len(var_list) == 1:
                return [shapes, ()]
            else:
                return [shapes, (len(var_list))]
        else:
            return shapes

    if isinstance(datafield_list, DataSchema):
        return data_schema2shapes(datafield_list)
    elif isinstance(datafield_list, list):
        shapes = []
        for field in datafield_list:
            if field.processor is None:
                continue
            shape = data_schema2shapes(field)
            if isinstance(shape, list):
                shapes.extend(shape)
            else:
                shapes.append(shape)
        return shapes

# 返回可变维度的下标的列表，用于生成可变维度的长度数据使用


def get_variable_shape_index(shape):
    num = []
    if isinstance(shape, (tuple, list)):
        for i in range(len(shape)):
            if shape[i] is None:
                num.append(i)
            elif not isinstance(shape[i], int):
                return None
    else:
        if shape is None:
            return [0]
    return num

# 判断dataschemas是否存在变长情况


def is_batch_padding(shape):
    if isinstance(shape, (tuple, list)):
        for sub_shape in shape:
            if is_batch_padding(sub_shape):
                return True
    else:
        return shape is None



#!/usr/bin/env python

"""Tests for `datasets` package.
    ref : https://docs.python.org/zh-cn/3/library/unittest.html
"""

import unittest
import cProfile
import os
import numpy as np
import tensorflow as tf

from datasets import TextlineParser, SeqParser
from datasets import TFDataset
from datasets.utils import TokenDicts, DataSchema


class TestDatasets(unittest.TestCase):
    """Tests for `datasets` package."""

    def setUp(self):
        if tf.__version__.startswith('1.'):
            tf.compat.v1.enable_eager_execution()
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def debug_tfrecord_format(self, tfrecord_file):
        filenames = [tfrecord_file]
        raw_dataset = tf.data.TFRecordDataset(filenames)
        for raw_record in raw_dataset.take(1):
            example = tf.train.Example()
            example.ParseFromString(raw_record.numpy())
            print(example)

    def pass_allway_dataset(self, generator, batch_size=4):
        print('is_training = True ', 'weight_fn=1')

        def weight_fn(x):
            return 1
        generator.parser.set_weight_fn(weight_fn)
        dataset = generator.generate_dataset(batch_size, 1, is_training=True)
        self.pass_dataset(True, True, dataset)
        print('is_training = False ', 'No weight_fn')
        generator.parser.set_weight_fn(None)
        dataset = generator.generate_dataset(batch_size, 1, is_training=False)
        self.pass_dataset(False, False, dataset)
        print('is_training = True ', 'No weight_fn')
        dataset = generator.generate_dataset(batch_size, 1, is_training=True)
        self.pass_dataset(True, False, dataset)

    def pass_dataset(self, is_training, weight_fn, dataset):
        if weight_fn:
            if is_training:
                for batch_num, (x, label, weight) in enumerate(dataset):
                    print('x', x)
                    print('weight', weight)
                    for d in x:
                        print('d.shape', d.shape)
                    print('label.shape', label.shape)
                    print('batch_num', batch_num)
                    break
        else:
            if is_training:
                for batch_num, (x, label) in enumerate(dataset):
                    print('x', x)
                    for d in x:
                        print('d.shape', d.shape)
                    print('label.shape', label.shape)
                    print('batch_num', batch_num)
                    break
            else:
                for batch_num, (info, x, label) in enumerate(dataset):
                    print('info', info)
                    print('x', x)
                    for d in x:
                        print('d.shape', d.shape)
                    print('label.shape', label.shape)
                    print('batch_num', batch_num)
                    break

    def test_seq_parser_dataset(self):
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        data_field_list.append(DataSchema(name='query', processor='to_tokenid',
                                          dtype='int32', shape=(None, None,), is_with_len=False, token_dict_name='query', has_session=True))
        data_field_list.append(DataSchema(name='width', processor='to_np', dtype='float32', shape=(None, 4,), has_session=True))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = SeqParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/seq_datasets', file_suffix='simple_seq.input')
        print('Shapes', generator._get_shapes(is_training=True))
        dataset = generator.generate_dataset(batch_size=1, num_epochs=1, is_shuffle=False)
        for data, i in zip(dataset.take(3), range(3)):
            (q, d), label = data
            if i == 0:
                q_y = np.array([[[68, 17, 33, 24]]], dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[1.]], dtype='float32')
            elif i == 1:
                q_y = np.array([[[68, 17, 33, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                 [2, 82, 9, 8, 3, 11, 69, 73, 15, 5, 46, 62, 99, 35]]],
                               dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[0.]], dtype='float32')
            elif i == 2:
                q_y = np.array([[[68, 17, 33, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                 [2, 82, 9, 8, 3, 11, 69, 73, 15, 5, 46, 62, 99, 35],
                                 [95, 34, 27, 1, 20, 74, 4, 20, 21, 3, 54, 57, 31, 29]]],
                               dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[2.]], dtype='float32')
            np.testing.assert_array_equal(q, q_y)
            np.testing.assert_array_equal(d, d_y)
            np.testing.assert_array_equal(label, label_y)

    def test_fast_seq_parser_dataset(self):
        # init token_dicts
        def weight(_):
            return 1.2
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        data_field_list.append(DataSchema(name='query', processor='to_tokenid',
                                          dtype='int32', shape=(None, None,), is_with_len=False, token_dict_name='query', has_session=True))
        data_field_list.append(DataSchema(name='width', processor='to_np', dtype='float32', shape=(None, 4,), has_session=True))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = SeqParser(token_dicts, data_field_list, label_field, weight_fn=weight)
        generator = TFDataset(parser=parser, file_path='tests/data/seq_datasets', file_suffix='simple_seq.input')
        print('Shapes', generator._get_shapes(is_training=True))
        dataset = generator.generate_dataset(
            batch_size=1, num_epochs=1, is_shuffle=False)
        for data, i in zip(dataset.take(3), range(3)):
            (q, d), label, weight = data
            if i == 0:
                q_y = np.array([[[68, 17, 33, 24]]], dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[1.]], dtype='float32')
            elif i == 1:
                q_y = np.array([[[68, 17, 33, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                 [2, 82, 9, 8, 3, 11, 69, 73, 15, 5, 46, 62, 99, 35]]],
                               dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[0.]], dtype='float32')
            elif i == 2:
                q_y = np.array([[[68, 17, 33, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                 [2, 82, 9, 8, 3, 11, 69, 73, 15, 5, 46, 62, 99, 35],
                                 [95, 34, 27, 1, 20, 74, 4, 20, 21, 3, 54, 57, 31, 29]]],
                               dtype='int32')
                d_y = np.array([[[0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28],
                                 [0.12, 0.34, 0.87, 0.28]]], dtype='float32')
                label_y = np.array([[2.]], dtype='float32')
            np.testing.assert_array_equal(q, q_y)
            np.testing.assert_array_equal(d, d_y)
            np.testing.assert_array_equal(label, label_y)

    def test_raw_query_float_dataset(self):
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        data_field_list.append(DataSchema(name='query', processor='to_tokenid',
                                          dtype='int32', shape=(None,), is_with_len=True, token_dict_name='query'))
        data_field_list.append(DataSchema(name='width', processor='to_np', dtype='float32', shape=(4)))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/raw_datasets', file_suffix='query_float.input')
        dataset = generator.generate_dataset(
            batch_size=2, num_epochs=1, is_shuffle=False)

        for d, i in zip(dataset.take(3), range(3)):
            (query, l, data), label = d
            if i == 0:
                query_y = np.array([[68, 17, 33, 24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                    [2, 82, 9, 8, 3, 11, 69, 73, 15, 5, 46, 62, 99, 35]],
                                   dtype='int32')
                l_y = np.array([4, 14], dtype='int32')
                data_y = np.array([[0.12, 0.34, 0.87, 0.28], [0.12, 0.34, 0.87, 0.28]], dtype='float32')
                label_y = np.array([[1.], [0.]], dtype='float32')
            elif i == 1:
                query_y = np.array([[95, 34, 27, 1, 20, 74, 4, 20, 21, 3, 54, 57, 31, 29],
                                    [78, 94, 19, 30, 26, 25, 0, 0, 0, 0, 0, 0, 0, 0]], dtype='int32')
                l_y = np.array([14, 6], dtype='int32')
                data_y = np.array([[0.12, 0.34, 0.87, 0.28], [0.12, 0.34, 0.87, 0.28]], dtype='float32')
                label_y = np.array([[2.], [1.]], dtype='float32')
            elif i == 2:
                query_y = np.array([[44, 18, 100, 84, 72, 59, 45, 63, 71, 28],
                                    [9, 8, 3, 11, 66, 5, 22, 35, 0, 0]], dtype='int32')
                l_y = np.array([10, 8], dtype='int32')
                data_y = np.array([[0.12, 0.34, 0.87, 0.28], [0.12, 0.34, 0.87, 0.28]], dtype='float32')
                label_y = np.array([[1.], [1.]], dtype='float32')
            np.testing.assert_array_equal(query, query_y)
            np.testing.assert_array_equal(l, l_y)
            np.testing.assert_array_equal(data, data_y)
            np.testing.assert_array_equal(label, label_y)

    def test_raw_dataset_varnum(self):
        token_dicts = None
        data_field_list = []
        data_field_list.append(DataSchema(name='query', processor='to_np',
                                          dtype='int32', shape=(None,), is_with_len=True))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,), is_with_len=False)
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/raw_datasets', file_suffix='varnum.input')
        # self.pass_allway_dataset(generator, 4)
        dataset = generator.generate_dataset(batch_size=1, num_epochs=1)
        for d, i in zip(dataset.take(3), range(3)):
            (q, l), label = d
            if i == 0:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 435, 234, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 1:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 4, 2, 9, 4, 5, 6, 2, 4]], dtype='int32')
                l_y = np.array([14], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 2:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 45, 24, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')

            np.testing.assert_array_equal(q, q_y)
            np.testing.assert_array_equal(l, l_y)
            np.testing.assert_array_equal(label, label_y)

    def test_text_seq2seq_model(self):
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        data_field_list.append(DataSchema(name='query', processor='to_tokenid',
                                          dtype='int32', shape=(None,), is_with_len=False, token_dict_name='query'))
        label_field = DataSchema(
            name='label',
            processor='to_tokenid',
            dtype='int32',
            shape=(
                None,
            ),
            is_with_len=True,
            token_dict_name='query')
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/raw_datasets', file_suffix='text_seq2seq.input')
        dataset = generator.generate_dataset(batch_size=12, num_epochs=1, is_shuffle=False)
        # for (batchs, (inputs, targets)) in enumerate(dataset):
        #    print('bacths', batchs, 'inputs', inputs, 'targets', targets)

    def test_tfrecord_dataset_varnum_writer_and_reader(self):
        token_dicts = None
        data_field_list = []
        data_field_list.append(DataSchema(name='query', processor='to_np', dtype='int32', shape=(None,), is_with_len=True))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,), is_with_len=False)
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/raw_datasets', file_suffix='varnum.input')
        if not os.path.exists('outputs'):
            os.mkdir('outputs')
        generator.to_tfrecords('outputs/file.tfrecord')
        generator = TFDataset(parser=parser, file_path='outputs', file_suffix='file.tfrecord', file_system='tfrecord')
        dataset = generator.generate_dataset(batch_size=1, num_epochs=1)
        for d, i in zip(dataset.take(3), range(3)):
            (q, l), label = d
            if i == 0:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 435, 234, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 1:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 4, 2, 9, 4, 5, 6, 2, 4]], dtype='int32')
                l_y = np.array([14], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 2:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 45, 24, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')

            np.testing.assert_array_equal(q, q_y)
            np.testing.assert_array_equal(l, l_y)
            np.testing.assert_array_equal(label, label_y)

        def weight(_):
            return 1.2
        parser = TextlineParser(token_dicts, data_field_list, label_field, weight_fn=weight)
        generator = TFDataset(parser=parser, file_path='tests/data/raw_datasets', file_suffix='varnum.input')
        if not os.path.exists('outputs'):
            os.mkdir('outputs')
        generator.to_tfrecords('outputs/file_weight.tfrecord')
        generator = TFDataset(parser=parser, file_path='outputs', file_suffix='file_weight.tfrecord', file_system='tfrecord')
        dataset = generator.generate_dataset(batch_size=1, num_epochs=1)
        for d, i in zip(dataset.take(3), range(3)):
            (q, l), label, w = d
            if i == 0:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 435, 234, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 1:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 4, 2, 9, 4, 5, 6, 2, 4]], dtype='int32')
                l_y = np.array([14], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')
            elif i == 2:
                q_y = np.array([[2, 3, 4, 6, 8, 23, 45, 24, 12, 234, 234]], dtype='int32')
                l_y = np.array([11], dtype='int32')
                label_y = np.array([[2.]], dtype='float32')

            self.assertAlmostEqual(w, 1.2)
            np.testing.assert_array_equal(q, q_y)
            np.testing.assert_array_equal(l, l_y)
            np.testing.assert_array_equal(label, label_y)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""Tests for `datasets` package.
    ref : https://docs.python.org/zh-cn/3/library/unittest.html
    ref : assert断言 https://www.jianshu.com/p/eea0b0e432da
"""

import unittest
import numpy as np
import tensorflow as tf

from datasets.utils import DataSchema
from datasets import TextlineParser


class TestDatasets(unittest.TestCase):
    """Tests for `datasets` package."""

    def setUp(self):
        if tf.__version__.startswith('1.'):
            tf.compat.v1.enable_eager_execution()
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_simple_no_label_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np',
                                             dtype='int32', shape=(None,), is_with_len=True))
        parser = TextlineParser(token_dicts, feature_field_list)
        line = '12\t1 2 3 4 5'
        label, addinfo, features, _ = parser.parse(line)
        assert label is None
        self.assertListEqual(addinfo, ['12'])
        np.testing.assert_array_equal(features[0], np.asarray([1, 2, 3, 4, 5]))

    def test_simple_label_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np',
                                             dtype='int32', shape=(None,), is_with_len=False))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = TextlineParser(token_dicts, feature_field_list, label_field)
        line = '12\t2\t1 2 3 4 5'
        label, addinfo, features, _ = parser.parse(line)
        np.testing.assert_array_equal(label, np.asarray([2.0]))
        self.assertListEqual(addinfo, ['12'])
        np.testing.assert_array_equal(features, np.asarray([1, 2, 3, 4, 5]))

    def test_multi_feat_single_label_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np', dtype='int32', shape=(None,), is_with_len=True))
        feature_field_list.append(DataSchema(name='width', processor='to_np', dtype='int32', shape=(4)))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = TextlineParser(token_dicts, feature_field_list, label_field)
        line = '12\t2\t1 2 3 4 5\t1 2 3 4'
        label, addinfo, features, _ = parser.parse(line)
        # print('label',label, features)
        np.testing.assert_array_equal(label, np.asarray([2.0]))
        self.assertListEqual(addinfo, ['12'])
        np.testing.assert_array_equal(features[0], np.asarray([1, 2, 3, 4, 5]))
        np.testing.assert_array_equal(features[2], np.asarray([1, 2, 3, 4]))

    def test_multi_feat_multi_label_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np', dtype='int32', shape=(None,), is_with_len=True))
        feature_field_list.append(DataSchema(
            name='width', processor='to_np', dtype='int32', shape=(4)))
        label_field = [DataSchema(name='label1', processor='to_np', dtype='float32', shape=(1,)),
                       DataSchema(name='label2', processor='to_np', dtype='int32', shape=(None,), is_with_len=True)]

        parser = TextlineParser(token_dicts, feature_field_list, label_field, additive_schema=['id', 'session'])
        line = '12\tcuid\t2\t2 3 4\t1 2 3 4 5\t1 2 3 4'
        label, addinfo, features, _ = parser.parse(line)
        # print('label',label, features)
        assert len(label) == 3
        np.testing.assert_array_equal(label[0], np.asarray([2.0]))
        np.testing.assert_array_equal(label[1], np.asarray([2, 3, 4]))
        self.assertListEqual(addinfo, ['12', 'cuid'])
        assert len(features) == 3
        np.testing.assert_array_equal(features[0], np.asarray([1, 2, 3, 4, 5]))
        assert features[1] == 5
        np.testing.assert_array_equal(features[2], np.asarray([1, 2, 3, 4]))

    def test_multi_feat_none_multi_label_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np', dtype='int32', shape=(None,), is_with_len=True))
        feature_field_list.append(DataSchema(name='width', processor=None, dtype='int32', shape=(4)))
        label_field = [DataSchema(name='label1', processor='to_np', dtype='float32', shape=(1,)),
                       DataSchema(name='label2', processor='to_np', dtype='int32', shape=(None,), is_with_len=True)]

        parser = TextlineParser(token_dicts, feature_field_list, label_field, additive_schema=['id', 'session'])
        line = '12\tcuid\t2\t2 3 4\t1 2 3 4 5\t1 2 3 4'
        label, addinfo, features, _ = parser.parse(line)
        # print('label',label, features)
        assert len(label) == 3
        np.testing.assert_array_equal(label[0], np.asarray([2.0]))
        np.testing.assert_array_equal(label[1], np.asarray([2, 3, 4]))
        self.assertListEqual(addinfo, ['12', 'cuid'])
        assert len(features) == 2
        np.testing.assert_array_equal(features[0], np.asarray([1, 2, 3, 4, 5]))
        assert features[1] == 5

    def test_multi_feat_multi_label_none_parser(self):
        # init token_dicts
        token_dicts = None
        feature_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        feature_field_list.append(DataSchema(name='query', processor='to_np', dtype='int32', shape=(None,), is_with_len=True))
        feature_field_list.append(DataSchema(name='width', processor='to_np', dtype='int32', shape=(4)))
        label_field = [DataSchema(name='label1', processor='to_np', dtype='float32', shape=(1,)),
                       DataSchema(name='label2', processor=None, dtype='int32', shape=(None,), is_with_len=True)]

        parser = TextlineParser(token_dicts, feature_field_list, label_field, additive_schema=['id', 'session'])
        line = '12\tcuid\t2\t2 3 4\t1 2 3 4 5\t1 2 3 4'
        label, addinfo, features, _ = parser.parse(line)
        # print('label',label, features)
        assert len(label) == 1
        np.testing.assert_array_equal(label, np.asarray([2.0]))
        self.assertListEqual(addinfo, ['12', 'cuid'])
        np.testing.assert_array_equal(features[0], np.asarray([1, 2, 3, 4, 5]))
        np.testing.assert_array_equal(features[2], np.asarray([1, 2, 3, 4]))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

"""Tests for `datasets` package."""

import unittest
# import sklearn
from datasets.utils.token_dicts import TokenDicts


def make_pkl_dict():
    dict_path = 'tests/data/dict/time.pkl'
    return dict_path


class TestDatasets(unittest.TestCase):
    """Tests for `datasets` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_vocab_dict_no_min_freq(self):
        """Test something."""
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        print('min_freq no', token_dicts.dict_min_freq)
        print('unkown no', token_dicts.unkown_emb_id)
        id_find = token_dicts.to_id('query', '月')
        assert id_find != token_dicts.unkown_emb_id['query']
        print('find 月:', id_find)
        id_find = token_dicts.to_id('query', '在')
        assert id_find != token_dicts.unkown_emb_id['query']
        print('find 在:', id_find)
        id_find = token_dicts.to_id('query', '101')
        assert id_find == token_dicts.unkown_emb_id['query']
        print('find 101:', id_find)

    def test_vocab_dict_min_freq_float(self):
        """Test something."""
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0.5})
        print('min_freq float', token_dicts.dict_min_freq)

    def test_vocab_dict_min_freq_int(self):
        """Test something."""
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'query': 2})
        print('min_freq int', token_dicts.dict_min_freq)
        print('unkown int', token_dicts.unkown_emb_id)
        id_find = token_dicts.to_id('query', '月')
        assert id_find == token_dicts.unkown_emb_id['query']
        print('find 月:', id_find)
        id_find = token_dicts.to_id('query', '在')
        assert id_find != token_dicts.unkown_emb_id['query']
        print('find 在:', id_find)
        id_find = token_dicts.to_id('query', '101')
        assert id_find == token_dicts.unkown_emb_id['query']
        print('find 101:', id_find)


"""
    def test_pkl_dict(self):
        # init token_dicts
        token_dicts = TokenDicts('tests/data/dicts', {'time': 'pkl'})
"""

if __name__ == '__main__':
    unittest.main()



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""

"""
from __future__ import absolute_import, division, print_function
import codecs


class TokenDicts(object):
    """
        处理数据转换成embedding id的类，所有的padding zero的token都为0。例如：
        1）text to embedding id ，通过dict实现
        2）连续值(包括向量和标量)离散化 to embedding id,　通过SKLEARN的kmeans来实现
        token dict数据组织：
        dict_path/dict_name.vocab : for 1)
        dict_path/dict_name.pkl : for 2)
    """

    def __init__(self, dict_path, dicts_param, search_range=100):
        # dicts_param : {dict_name, min_freq}
        # min_freq > 1　则是直接设定min_freq;  <1 :则按照百分比决定min_freq
        # TODO : 统一 names_dicts min_freq_dicts到一个输入
        # TODO : 通过cutoff百分比来确定min_freq参数，即去除percent的freq占比，可以取0.05
        self.dicts = {}
        self.dict_size = {}
        self.unkown_emb_id = {}
        self.dict_min_freq = {}
        self.search_range = search_range
        for (dict_name, min_freq) in dicts_param.items():
            if isinstance(min_freq, int):
                self.dict_min_freq[dict_name] = float(min_freq)
            elif isinstance(min_freq, float):
                if min_freq >= 1.0:
                    self.dict_min_freq[dict_name] = min_freq
        for (dict_name, min_freq) in dicts_param.items():
            if min_freq == 'pkl':
                dict_file = dict_path + '/' + dict_name + '.pkl'
                self._vector_to_emb_id(dict_file, dict_name)
            elif isinstance(min_freq, int) or isinstance(min_freq, float):
                dict_file = dict_path + '/' + dict_name + '.vocab'
                if dict_name not in self.dict_min_freq:
                    self.dict_min_freq[dict_name] = self._compute_min_freq_by_percent(dict_file, min_freq)
                self._text_to_emb_id(dict_file, dict_name)

    def _compute_min_freq_by_percent(self, filename, percent):
        print('percent', percent)
        search_min_freqs = [0 for i in range(1, self.search_range)]
        sum_freqs = 0.0
        with codecs.open(filename, encoding='utf8') as f:
            for line in f:
                _, freq = line.strip('\n').split('\t')
                freq = float(freq)
                sum_freqs += freq
                for i in range(0, len(search_min_freqs)):
                    if freq > i + 1:
                        search_min_freqs[i] += freq
        for i in range(0, len(search_min_freqs)):
            if search_min_freqs[i] < (1 - percent) * sum_freqs:
                return i
        return self.search_range

    def _vector_to_emb_id(self, filename, dict_name):
        token_dict = {}
        token_id = 0
        for line in codecs.open(filename, encoding='utf8'):
            token, freq = line.strip('\n').split('\t')
            freq = int(freq)
            if freq >= self.dict_min_freq[dict_name]:
                token_id = token_id + 1
                token_dict[token] = token_id
        self.dicts[dict_name] = token_dict
        self.unkown_emb_id[dict_name] = token_id + 1
        self.dict_size[dict_name] = token_id + 2

    def _text_to_emb_id(self, filename, dict_name):
        token_dict = {}
        token_id = 0
        with codecs.open(filename, encoding='utf8') as f:
            for line in f:
                token, freq = line.strip('\n').split('\t')
                freq = float(freq)
                if freq >= self.dict_min_freq[dict_name]:
                    token_id = token_id + 1
                    token_dict[token] = token_id
        self.dicts[dict_name] = token_dict
        self.unkown_emb_id[dict_name] = token_id + 1
        self.dict_size[dict_name] = token_id + 2

    def dict_size_by_name(self, dict_name):
        """
        返回字典大小
        Args:
            dict_name: 名称

        Returns:
            int类型 dict大小
        """
        return self.dict_size[dict_name]

    def to_id(self, dict_name, token):
        """
        token to id
        Args:
            dict_name: 名称
            token: 分词

        Returns:
            int类型 token id
        """
        token_dict = self.dicts[dict_name]
        if token in token_dict:
            return token_dict[token]
        else:
            return self.unkown_emb_id[dict_name]



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import, division, print_function
import numpy as np
import soundfile as sf
import python_speech_features

__all__ = ['data_processor_dicts']


def to_np(x, dtype):
    x_data = [i if i != 'None' else 0 for i in x.split(' ')]
    return np.asarray(x_data, dtype=dtype)


def to_tokenid(x, dict_name, token_dicts):
    x_data = [token_dicts.to_id(dict_name, token) for token in x.split(' ')]
    return np.asarray(x_data, dtype='int32')

# ref:https://docs.scipy.org/doc/numpy/reference/generated/numpy.pad.html
# @ TODO fix pad_width


def padding_seq(tokens, pad_width, pad_suffix=True, mode='constant', **kwargs):
    if len(tokens) >= pad_width:
        return tokens[0:pad_width]
    padding_num = pad_width - len(tokens)
    if pad_suffix:
        return np.pad(tokens, (0, padding_num), mode, **kwargs)
    else:
        return np.pad(tokens, (padding_num, 0), mode, **kwargs)


def to_fe_mfcc(x):
    """
    ref : https://python-speech-features.readthedocs.io/en/latest/
    Compute MFCC features from an audio signal.
    @param x: wav file
    return mfcc: shape[NUMFRAMES,numcep]
    """
    signal, samplerate = sf.read(x)
    mfcc = python_speech_features.base.mfcc(signal, samplerate=samplerate, winlen=0.025, winstep=0.01, numcep=13, nfilt=26, nffft=512, lowfreq=0,
                                            highfreq=None, preemph=0.97, ceplifter=22, appendEnergy=True, winfunc=np.hanmming)
    return mfcc


def to_fe_fbank(x):
    """
    Compute Mel-filterbank energy features from an audio signal.
    @param x: wav file
    return fbank: shape[NUMFRAMES, nfilt]
    """
    signal, samplerate = sf.read(x)
    fbank, _ = python_speech_features.base.fbank(signal, samplerate=samplerate, winlen=0.025, winstep=0.01, nfilt=26, nffft=512, lowfreq=0,
                                                 highfreq=None, preemph=0.97, winfunc=np.hanmming)
    return fbank


def to_fe_logfbank(x):
    """
    Compute log Mel-filterbank energy features from an audio signal.
    @param x: wav file
    return logfbank: shape[NUMFRAMES, nfilt]
    """
    signal, samplerate = sf.read(x)
    logfbank = python_speech_features.base.logfbank(signal, samplerate=samplerate, winlen=0.025, winstep=0.01, nfilt=26, nffft=512, lowfreq=0,
                                                    highfreq=None, preemph=0.97)
    return logfbank


def to_fe_ssc(x):
    """
    Compute Spectral Subband Centroid features from an audio signal
    @param x: wav file
    return logfbank: shape[NUMFRAMES, nfilt]
    """
    signal, samplerate = sf.read(x)
    logfbank = python_speech_features.base.logfbank(signal, samplerate=samplerate, winlen=0.025, winstep=0.01, nfilt=26, nffft=512, lowfreq=0,
                                                    highfreq=None, preemph=0.97)
    return logfbank


def token_process(token_id, vocab_size, token_mask_id, mask_prob=0.8, keep_prob=0.1):
    """以mask_prob的概率替换为[MASK]，以keep_prob的概率保持不变，否则替换为一个随机token
    """
    rand = np.random.random()
    if rand <= mask_prob:
        return token_mask_id
    elif rand <= mask_prob + keep_prob:
        return token_id
    else:
        return np.random.randint(1, vocab_size)  # 0 is padding id


def gen_causal_attention_mask(x):
    """
    下三角为1的矩阵
    e.g. x=np.ones([32,6,128])
    mask =
    array([[[[1, 0, 0, 0, 0, 0],
                    [1, 1, 0, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 1, 0, 0],
                    [1, 1, 1, 1, 1, 0],
                    [1, 1, 1, 1, 1, 1]]]], dtype=int32)
    """
    seq_len = x.shape[1]
    idxs = np.arange(0, seq_len)
    mask = idxs[None, :] <= idxs[:, None]
    mask = np.asarray(mask, 'int32')
    return mask[None, None]


def gen_unilm_attention_mask(x):
    """
    (UNILM)Unified Language Model Pre-training for Natural Language Understanding and Generation
    论文地址：https://arxiv.org/abs/1905.03197
    论文代码：https://github.com/microsoft/unilm
    next sentence :
    >>> x=np.asarray([0,0,0,1,1,1])
    >>> y=gen_unilm_attention_mask(x[None,:])
    >>> y
    array([[[[1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 0, 0, 0],
                    [1, 1, 1, 1, 0, 0],
                    [1, 1, 1, 1, 1, 0],
                    [1, 1, 1, 1, 1, 1]]]], dtype=int32)
    """
    idxs = np.cumsum(x, axis=1)
    mask = idxs[:, None, :] <= idxs[:, :, None]
    mask = np.asarray(mask, 'int32')
    return mask[:, None]


data_processor_dicts = {
    'to_np': lambda x, schema, token_dicts: to_np(x, schema.dtype),
    'to_tokenid': lambda x, schema, token_dicts: to_tokenid(x, schema.token_dict_name, token_dicts),
    'to_tokenid_start': lambda x, schema, token_dicts: to_tokenid('<s> ' + x, schema.token_dict_name, token_dicts),
    'to_tokenid_end': lambda x, schema, token_dicts: to_tokenid(x + r' <\s>', schema.token_dict_name, token_dicts),
    'to_sentenceid': lambda x, schema, token_dicts: to_tokenid('<s> ' + x + r' <\s>', schema.token_dict_name, token_dicts),
    'to_fe_mfcc': lambda x, _, __: to_fe_mfcc(x),
    'to_fe_fbank': lambda x, _, __: to_fe_fbank(x),
    'to_fe_logfbank': lambda x, _, __: to_fe_logfbank(x),
    'to_fe_ssc': lambda x, _, __: to_fe_ssc(x),
    'None': None
}



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""raw multi-inputs dataset generator class
"""

from __future__ import absolute_import, division, print_function
import numpy as np

from datasets.utils import data_processor_dicts
from datasets.parser import Parser


class TextlineParser(Parser):
    """ parser
    """
    # def __init__(self, data_field_list, label_field, additive_schema=['id']):
    #    self.

    def parse(self, line):
        label = None
        additive_info = []
        additive_info_len = len(self.additive_schema)
        if additive_info_len > 0:
            items = line.strip('\n').split('\t', additive_info_len)
            for i in range(0, additive_info_len):
                additive_info.append(items[i])
            line = items[-1]
        data_fields = line.strip('\n').split('\t')
        data_num_fields = len(data_fields)
        assert data_num_fields == len(self.data_field_list), 'line items:%d must equal fields num %d' % (data_num_fields, len(self.data_field_list))
        if data_num_fields == 0:
            raise ValueError("data_fields with size 0")
        # 根据data_field_list的token_dicts和processor来解析一行数据，如processor为None,则丢弃该数据
        datas = [data_processor_dicts[field.processor](text, field, self.token_dicts)
                 for text, field in zip(data_fields, self.data_field_list) if field.processor is not None]
        if self.is_data_padding:
            out_datas = []
            for feature, var_shape_index in zip(datas, self.var_shape_index_list):
                out_datas.append(feature)
                if var_shape_index is None:
                    continue
                if len(var_shape_index) == 1:
                    out_datas.append(
                        feature.shape[var_shape_index[0]])
                elif len(var_shape_index) > 1:
                    out_datas.append(np.asarray(
                        [feature.shape[k] for k in var_shape_index], dtype='int32'))
            datas = out_datas
        if self.label_range == 1:
            label = datas[0]
            features = datas[1:]
        elif self.label_range > 1:
            label = datas[:self.label_range]
            features = datas[self.label_range:]
        else:
            features = datas
        if len(features) == 1:
            features = features[0]
        if self.weight_fn is None:
            weight = None
        else:
            weight = self.weight_fn(line)
        return label, additive_info, features, weight



"""Top-level package for datasets."""

__author__ = """yinochaos"""
__email__ = 'pspcxl@163.com'
__version__ = '0.0.16'

__all__ = ['DataSchema', 'data_processor_dicts', 'TokenDicts', 'TextlineParser', 'SeqParser', 'TFDataset']
#__all__ = ['DataSchema', 'data_processor_dicts', 'TokenDicts', 'TextlineParser', 'SeqParser', 'TFDataset', 'PTDataset']
from datasets.textline_parser import TextlineParser
from datasets.seq_parser import SeqParser
from datasets.tf_dataset import TFDataset
#from datasets.pt_dataset import PTDataset
from datasets.utils import DataSchema, data_processor_dicts, TokenDicts



#!/usr/bin/env python

"""Tests for `datasets` package.
    ref : https://docs.python.org/zh-cn/3/library/unittest.html
"""

import unittest
import cProfile
import os
import tensorflow as tf

from datasets import TextlineParser, SeqParser
from datasets import TFDataset
from datasets.utils import TokenDicts, DataSchema

def run_tf_dataset_session_parser():
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        # param = ["name", "processor", "type", "dtype", "shape", "max_len", "token_dict_name"]
        data_field_list.append(DataSchema(name='query', processor='to_tokenid',
                                          dtype='int32', shape=(None, None,), is_with_len=False, token_dict_name='query', has_session=True))
        data_field_list.append(DataSchema(name='width', processor='to_np', dtype='float32', shape=(None, 4,), has_session=True))
        label_field = DataSchema(name='label', processor='to_np', dtype='float32', shape=(1,))
        parser = SeqParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path='tests/data/seq_datasets', file_suffix='simple_seq.input')
        #print('Shapes', generator._get_shapes(is_training=True))
        dataset = generator.generate_dataset(
            batch_size=12, num_epochs=10, is_shuffle=False)
        for _ in enumerate(dataset):
            pass

if __name__ == '__main__':
    #cProfile.run('run_tf_dataset_session_parser()')
    run_tf_dataset_session_parser()


#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.md') as history_file:
#    history = history_file.read()

# install requirements
requirements = ['scipy', 'tdqm', 'python_speech_features', 'numpy', 'soundfile']

setup_requirements = []

test_requirements = []

setup(
    author="yinochaos",
    author_email='pspcxl@163.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="datasets for easy machine learning use",
    entry_points={
        'console_scripts': [
            'datasets=datasets.cli:main',
        ],
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='ml-dataset',
    name='ml-dataset',
    packages=find_packages(include=['datasets', 'datasets.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/yinochaos/datasets',
    version='0.0.16',
    zip_safe=False,
)



"""Console script for datasets."""
import argparse
import sys
import os
from functools import wraps
from concurrent.futures import ThreadPoolExecutor
import tensorflow as tf
import numpy as np

from datasets.utils import DataSchema
from datasets import TFDataset, TextlineParser
from datasets.utils.common_struct import load_hdfs_filelist, load_local_filelist

# color print ref : https://blog.csdn.net/qq_34857250/article/details/79673698


def loop_input(a_func):
    @wraps(a_func)
    def decorated(*args, **kwargs):
        while True:
            result, msg = a_func(*args, **kwargs)
            if result is not None:
                return result
            else:
                print('\033[1;31m' + msg + '\033[0m')
    return decorated


@loop_input
def filepath_input(msg):
    filepath = input('请输入' + msg + '的文件路径:')
    filepath = filepath.strip()
    if os.path.exists(filepath):
        return filepath, None
    else:
        return None, filepath + 'is not find'


@loop_input
def shape_input():
    line = input('请输入数据的shape[e.g. 15,3]:')
    try:
        shape = [float(x) for x in line.split(',')]
        return shape, None
    except ValueError:
        return None, '输入格式错误，请输入各个维度，按照,隔开 e.g. 15,3'


@loop_input
def choose_input(choose_list, type_text):
    print(' '.join(choose_list))
    line = input(' 请从以上选项选出' + type_text + ':')
    if line in choose_list:
        return line, None
    else:
        return None, '请从选项当中选择，可以copy，请勿超出选项范围'


@loop_input
def name_input(msg):
    line = input(msg)
    line = line.strip()
    if len(line) > 0:
        return line, None
    else:
        return None, '请输入非空字符'


@loop_input
def yesno_input(msg, default='y'):
    if default == 'y':
        tail = '[Y/n]:'
        result = True
    else:
        tail = '[y/N]:'
        result = False
    line = input(msg + tail)
    line = line.strip()
    if line == '':
        return result
    elif line in {'yes', 'YES', 'Yes', 'y', 'Y'}:
        return True
    elif line in {'no', 'NO', 'No', 'n', 'N'}:
        return False
    else:
        return None, '请填入如下选项: y,n'

# def choose_processor_input():


def convert2tfrecords(dataset_file_path, file_suffix, tfrecord_file_path, n_workers):
    filenames = os.listdir(dataset_file_path)
    if file_suffix is None:
        filenames = [f for f in list(filenames)]
    else:
        filenames = [f for f in list(filenames) if f.endswith(file_suffix)]

    def processor(filepath, filename, tfrecord_filename):
        token_dicts = None
        data_field_list = []
        data_field_list.append(DataSchema(name='query', processor='to_np', type=tf.int32,
                                          dtype='int32', shape=(None,), is_with_len=True))
        label_field = DataSchema(name='label', processor='to_np',
                                 type=tf.float32, dtype='float32', shape=(1,), is_with_len=False)
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path=filepath, file_suffix=filename)
        generator.to_tfrecords(tfrecord_filename)
        return tfrecord_filename

    task_param_list = [tuple(dataset_file_path, filename, tfrecord_file_path + '/' + str(i) + '.tfrecord')
                       for filename, i in zip(filenames, len(filenames))]
    pool = ThreadPoolExecutor(max_workers=n_workers)
    for result in pool.map(processor, task_param_list):
        print(result, 'finish')


class CodeGenerator(object):
    def __init__(self):
        self.schema_list = []
        self.add_info = []
        self.label_len = 1
        self.feat_flags = []
        self.label_names = []
        self.feature_names = []
        self.init_code = 'batch_size num_epochs is_shuffle file_path file_suffix'
        self.code = """
        token_dicts = TokenDicts('tests/data/dicts', {'query': 0})
        data_field_list = []
        data_field_list.append(DataSchema(name='query', processor='to_tokenid', type=tf.int32,
                                          dtype='int32', shape=(None,), is_with_len=True, token_dict_name='query'))
        data_field_list.append(DataSchema(
            name='width', processor='to_np', type=tf.float32, dtype='float32', shape=(4)))
        label_field = DataSchema(
            name='label', processor='to_np', type=tf.float32, dtype='float32', shape=(1,))
        parser = TextlineParser(token_dicts, data_field_list, label_field)
        generator = TFDataset(parser=parser, file_path=file_path, file_suffix=file_suffix)
        dataset = generator.generate_dataset(
            batch_size=batch_size, num_epochs=num_epochs, is_shuffle=is_shuffle)
        for _ in enumerate(dataset):
            pass
        """
        self.convert_dicts = {
            'wav': ['to_fe_mfcc', 'audio', 'float', [None, 13]],
            'pcm': ['to_fe_mfcc', 'audio', 'float', [None, 13]],
            'mp3': ['to_fe_mfcc', 'audio', 'float', [None, 13]],
            'jpg': ['img_read', 'img', 'float', None],
            'png': ['img_read', 'img', 'float', None],
            'JPG': ['img_read', 'img', 'float', None]
        }

    def _get_field_processor(self, colum_list):
        is_same_shape = True
        processor, type, dtype, shape = self._data_field_check(colum_list[0])
        for data, index in zip(colum_list, range(len(colum_list))):
            types = self._data_field_check(data)
            if processor != types[0]:
                print(processor, ' not match line', index, colum_list[index], types[0])
                sys.exit(-1)
            if type != types[1]:
                print(type, ' not match line', index, colum_list[index], types[1])
                sys.exit(-1)
            if dtype != types[2]:
                print(dtype, ' not match line', index, colum_list[index], types[2])
                sys.exit(-1)
            if shape != types[3]:
                is_same_shape = False
        if shape is None:
            shape = shape_input()
        if processor in {'to_np', 'to_tokenid'} and not is_same_shape:
            shape = [None]
        print('processor', processor, 'type', type, 'dtype', dtype, 'shape', shape, 'for data colums')
        print(colum_list[:3])
        return [processor, type, dtype, shape]

    def _write_token_dicts(self, f, label_schema, features_schema):
        dict_names = set()
        name_dicts = {}
        for name, schema in zip(self.label_names, label_schema):
            if schema[0] in {'to_tokenid'}:
                dict_name = name_input('input dict name for' + name + ' :')
                dict_names.add(dict_name)
                name_dicts[name] = dict_name
        for name, schema in zip(self.feature_names, features_schema):
            if schema[0] in {'to_tokenid'}:
                dict_name = name_input('input dict name for feature :' + name + ' :')
                dict_names.add(dict_name)
                name_dicts[name] = dict_name
        if len(dict_names) == 0:
            f.write('token_dicts = None\n')
        else:
            name_list = '{' + ','.join(["'%s':0" % (x) for x in dict_names]) + '}'
            f.write('token_dicts = TokenDicts(\'dicts\', ' + name_list + ')\n')
        # TODO 增加具体的dict生成的处理
        return name_dicts

    def _write_schema_list(self, f, schema_list, name_list, list_name, name_dicts):
        f.write(list_name + ' = []\n')
        # data_field_list.append(DataSchema(
        #    name='width', processor='to_np', type=tf.float32, dtype='float32', shape=(4)))
        for name, schema in zip(name_list, schema_list):
            shape_text = ','.join([x if x else 'None' for x in schema[3]])
            if shape_text == 'None':
                shape_text = 'None,'
            if name in name_dicts:
                f.write("%s.append(DataSchema(name='%s', processor='%s', dtype='%s', shape=(%s), token_dict_name='%s'))\n" %
                        (list_name, name, schema[0], schema[2], shape_text, name_dicts[name]))
            else:
                f.write("%s.append(DataSchema(name='%s', processor='%s', dtype='%s', shape=(%s)))\n" %
                        (list_name, name, schema[0], schema[2], shape_text))

    def generate_code(self):
        workspace = name_input('输入代码生成目录 : ')
        if not os.path.exists(workspace):
            os.mkdir(workspace)
        code_file = workspace + '/dataset_reader.py'
        data_file = filepath_input('输入数据文件')
        file_suffix = input('输入数据文件后缀 : ')
        file_suffix = file_suffix.strip()
        file_list = load_local_filelist(data_file, file_suffix)
        assert len(file_list) > 0, "in %s can't find file with %s" % (data_file, file_suffix)
        lines = []
        with open(file_list[0], 'r') as f:
            for line in f:
                lines.append(line.strip('\n'))
                if len(lines) > 20:
                    break
        with open(code_file, 'w') as f:
            f.write('\n'.join(['import tensorflow as tf', 'from datasets import TextlineParser',
                               'from datasets import TFDataset', 'from datasets.utils import TokenDicts, DataSchema']))
            f.write('\n')
            f.write("file_path = '" + data_file + "'\n")
            if file_suffix == '':
                f.write("file_suffix = None\n")
            else:
                f.write("file_suffix = '" + file_suffix + "'\n")
            f.write("batch_size = 64\n")
            f.write("num_epochs = 1\n")
            f.write("is_shuffle = True\n")

            self._parser_header(lines[0])
            label_items = []
            feature_items = []
            print('lines', lines)
            for line in lines[1:]:
                items = line.split('\t')
                items = items[len(self.add_info):]
                label_items.append(items[0:self.label_len])
                feature_items.append(items[self.label_len:])
            label_items = np.asarray(label_items)
            feature_items = np.asarray(feature_items)
            print('items shape', label_items.shape)
            label_schema = []
            features_schema = []
            for colum in range(label_items.shape[1]):
                colum_list = label_items[:, colum].tolist()
                label_schema.append(self._get_field_processor(colum_list))
            for colum in range(feature_items.shape[1]):
                colum_list = label_items[:, colum].tolist()
                features_schema.append(self._get_field_processor(colum_list))
            print('add_info', self.add_info)
            print('label schema:', label_schema)
            print('feature schema:', features_schema)
            name_dicts = self._write_token_dicts(f, label_schema, features_schema)
            self._write_schema_list(f, label_schema, self.label_names, 'label_schema_list', name_dicts)
            self._write_schema_list(f, features_schema, self.feature_names, 'feature_schema_list', name_dicts)
            f.write("parser = TextlineParser(token_dicts, feature_schema_list, label_schema_list)\n")
            f.write("generator = TFDataset(parser=parser, file_path=file_path, file_suffix=file_suffix)\n")
            f.write("dataset = generator.generate_dataset(batch_size=batch_size, num_epochs=num_epochs, is_shuffle=is_shuffle)\n")
            f.write("for _ in enumerate(dataset):\n    pass\n")

    def _parser_header(self, head):
        items = head.split('\t')
        self.label_len = 0
        self.feat_flags = []
        i = 0
        while i < len(items):
            if items[i].startswith('label'):
                break
            else:
                self.add_info.append(items[i])
            i += 1
        while i < len(items):
            if items[i].startswith('label'):
                self.label_len += 1
                self.label_names.append(items[i])
            else:
                break
            i += 1
        while i < len(items):
            self.feature_names.append(items[i])
            if items[i].startswith('fe_'):
                self.feat_flags.append(True)
            else:
                self.feat_flags.append(False)
            i += 1

    def _data_field_check(self, text):
        file_suffix = text.split('.')[-1]
        if file_suffix in self.convert_dicts:
            return self.convert_dicts[file_suffix]
        items = text.split(' ')
        is_float = True
        is_int = True
        for item in items:
            if not item.isnumeric():
                is_int = False
            try:
                float(item)
            except ValueError:
                is_float = False
        if is_int:
            return ['to_np', 'class', 'int', len(items)]
        if is_float:
            return ['to_np', 'num', 'float', len(items)]
        return ['to_tokenid', 'text', 'int', len(items)]


def main():
    """Console script for datasets."""
    code_generator = CodeGenerator()
    code_generator.generate_code()
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""raw multi-inputs dataset generator class
"""

from __future__ import absolute_import, division, print_function
import os
import subprocess
import tensorflow as tf
import codecs
import pickle

from datasets.utils.common_struct import data_schemas2shapes, data_schemas2types, type2tf_dict
from datasets.utils.common_struct import load_hdfs_filelist, load_local_filelist


class TFDataset(object):
    """从文件流创建的dataset
        利用tf.dataset支持多文件输入（本地和HDFS同时支持），对于大规模数据集十分友好；
        并通过token_dicts和datafields支持配置化的数据处理，灵活支持多输入数据集的处理
        text（local ， hdfs）：当前已经支持
        @TODO : 支持多种文件读取方式
        pickle：finish TODO test
        qiuckle:
        tfrecord：code&test finish
        LMDB（Lightning Memory-Mapped Database(快如闪电的内存映射数据库)）：TODO
        HDF5：TODO
    """
    ##
    # @brief
    #
    # @param file_path dataset文件路径
    # @param token_dicts 分词id词典
    # @param data_field_list dataset的schema数据
    # @param label_field label的schema数据
    # @param is_with_label dataset是否带有label
    # @param file_system
    #               local  : 从本地读取文件
    #               HDFS :  从HDFS读取文件
    # @param file_suffix dataset文件后缀,如果是None则使用所有文件
    # @param weight_fn 计算样本权重的函数 weight_fn(line)
    #
    # @return

    def __init__(self, parser, file_path, file_system='local', file_suffix=None, filenames=None):
        self.file_system = file_system
        self.parser = parser
        self.check_schemalist(parser.feature_field + parser.label_field)
        if self.file_system == 'HDFS':
            self.hadoop = os.environ.get('HADOOP')
            if self.hadoop is None:
                raise RuntimeError(
                    'os.environ[HADOOP] is None, should export HADOOP')
        if filenames is None:
            self.filenames = self.load_filelist(file_path, file_suffix, file_system)
            print('file names', self.filenames)
        else:
            self.filenames = filenames
        self.file_suffix = file_suffix

        if len(data_schemas2types(self.parser.label_field)) > 1:
            self.is_multi_label = True
        else:
            self.is_multi_label = False

        if len(data_schemas2types(self.parser.feature_field)) > 1:
            self.is_multi_features = True
        else:
            self.is_multi_features = False

    def check_schemalist(self, schema_list):
        for schema in schema_list:
            if schema.dtype not in type2tf_dict:
                raise ValueError("schema  %s :%s is not valid" % (schema.name, schema.dtype))

    ##
    # @brief
    #   载入目录所有模型输入文件
    # @param datadir 数据目录
    # @param file_suffix 文件后缀
    # @param file_system
    #
    # @return filenames 所有文件列表

    def load_filelist(self, datadir, file_suffix, file_system):
        filenames = []
        if file_system == 'HDFS':
            filenames = load_hdfs_filelist(datadir, file_suffix, self.hadoop)
        else:
            filenames = load_local_filelist(datadir, file_suffix)
        return filenames

    ##
    # @brief
    #       file的读取迭代器，支持本地文件和HDFS文件的读取
    #       TODO ： 可以考虑添加其他网络文件的读取支持
    #    pickle：
    #    tfrecord：
    #    LMDB（Lightning Memory-Mapped Database(快如闪电的内存映射数据库)）：
    #    HDF5：
    # @param filename
    #
    # @return line
    def read_input_lines(self, filename):
        if self.file_system == 'HDFS':
            cmd = self.hadoop + ' fs -cat ' + filename
            with subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True) as pipe:
                for line in iter(pipe.stdout.readline, b''):
                    yield line.decode('utf8', 'ignore')
        elif self.file_system == 'local':
            with codecs.open(filename, encoding='utf8') as f:
                for line in f:
                    yield line
        elif self.file_system == 'pickle':
            with open(filename, 'rb') as f:
                while True:
                    try:
                        ex = pickle.load(f)
                        if isinstance(ex, list):
                            yield from ex
                        else:
                            yield ex
                    except EOFError:
                        break

    ##
    # @brief
    #       file的读取迭代器，支持本地文件和HDFS文件的读取
    #
    # @return dataset output

    def data_generator(self):
        """
        data_generator
        Returns:
            batch
        """
        for filename in self.filenames:
            for line in self.read_input_lines(filename):
                result = self.parser.parse(line)
                if result is None:
                    continue
                label, additive_info, features, _ = result
                if self.is_multi_label:
                    label = tuple(label)
                # output datas
                if self.parser.label_field is not None:
                    yield additive_info, tuple(features), label
                else:
                    yield additive_info, tuple(features)

    ##
    # @brief
    #       file的读取迭代器，支持本地文件和HDFS文件的读取
    #
    # @return dataset output
    def data_generator_training(self):
        """
        data_generator
        Returns:
            batch
        """
        for filename in self.filenames:
            for line in self.read_input_lines(filename):
                label, _, features, weight = self.parser.parse(line)
                if self.is_multi_label:
                    label = tuple(label)
                # output datas
                if weight is not None:
                    # 用于sample_weights的dataset: (inputs, targets, sample_weights)
                    if weight > 0:
                        yield tuple(features), label, weight
                else:
                    yield tuple(features), label

    ##
    # @brief
    #       file的读取迭代器，支持tfrecord
    #
    # @return dataset output

    def data_generator_tfrecord_flat(self):
        """
        data_generator
        Returns:
            batch
        """
        for filename in self.filenames:
            for line in self.read_input_lines(filename):
                label, _, features, weight = self.parser.parse(line)
                # output datas
                if weight is not None:
                    if weight > 0:
                        yield tuple(features + [label, weight])
                else:
                    yield tuple(features + [label])

    def _get_types(self, is_training, is_flat=False):
        types = []
        feature_types = data_schemas2types(self.parser.feature_field)
        label_types = None
        if self.parser.has_label:
            label_types = data_schemas2types(self.parser.label_field)
        if is_flat:
            types.extend(feature_types)
            if self.parser.has_label:
                types.extend(label_types)
            if self.parser.weight_fn is not None:
                types.append(tf.float32)
            return tuple(types)

        if len(feature_types) > 1:
            feature_types = tuple(feature_types)
        else:
            feature_types = feature_types[0]
        if len(label_types) > 1:
            label_types = tuple(label_types)
        else:
            label_types = label_types[0]
        if is_training:
            if self.parser.weight_fn is not None:
                types = tuple([feature_types, label_types, tf.float32])
            else:
                types = tuple([feature_types, label_types])
        else:
            if self.parser.has_label:
                types = tuple([tf.string, feature_types, label_types])
            else:
                types = tuple([tf.string, feature_types])
        return types

    def _get_shapes(self, is_training, is_flat=False):
        shapes = []
        feature_shapes = data_schemas2shapes(self.parser.feature_field)
        if self.parser.has_label:
            label_shapes = data_schemas2shapes(self.parser.label_field)
        else:
            label_shapes = None
        if is_flat:
            shapes.extend(feature_shapes)
            if self.parser.has_label:
                shapes.extend(label_shapes)
            if self.parser.weight_fn is not None:
                shapes.append(())
            return tuple(shapes)

        if len(feature_shapes) > 1:
            feature_shapes = tuple(feature_shapes)
        else:
            feature_shapes = feature_shapes[0]
        if len(label_shapes) > 1:
            label_shapes = tuple(label_shapes)
        else:
            label_shapes = label_shapes[0]
        if is_training:
            if self.parser.weight_fn is not None:
                shapes = tuple([feature_shapes, label_shapes, ()])
            else:
                shapes = tuple([feature_shapes, label_shapes])
        else:
            if self.parser.has_label:
                shapes = tuple([(len(self.parser.additive_schema)), feature_shapes, label_shapes])
            else:
                shapes = tuple([len(self.parser.additive_schema), feature_shapes])
        return shapes

    def _from_tfrecord(self):
        feature_description = {}
        for name in self.parser.features_dict:
            feature_description[name] = tf.io.FixedLenFeature([], tf.string)
            if self.parser.features_dict[name].is_with_len:
                feature_description[name + '_len'] = tf.io.FixedLenFeature([], tf.string)
        for name in self.parser.label_dict:
            feature_description[name] = tf.io.FixedLenFeature([], tf.string)
            if self.parser.label_dict[name].is_with_len:
                feature_description[name + '_len'] = tf.io.FixedLenFeature([], tf.string)
        if self.parser.weight_fn is not None:
            feature_description['weight'] = tf.io.FixedLenFeature([], tf.string)
        print('feature_des', feature_description)

        def _parse_function(example_proto):
            # Parse the input tf.train.Example proto using the dictionary above.
            data = tf.io.parse_single_example(example_proto, feature_description)
            features = []
            for field in self.parser.features_dict.values():
                features.append(tf.io.parse_tensor(data[field.name], type2tf_dict[field.dtype]))
                if field.is_with_len:
                    features.append(tf.io.parse_tensor(data[field.name + '_len'], tf.int32))
            if self.parser.has_label:
                weight = None
                if self.parser.weight_fn is not None:
                    weight = tf.io.parse_tensor(data['weight'], tf.float32)
                labels = []
                for field in self.parser.label_dict.values():
                    labels.append(tf.io.parse_tensor(data[field.name], type2tf_dict[field.dtype]))
                    if field.is_with_len:
                        labels.append(tf.io.parse_tensor(data[field.name + '_len'], tf.int32))
                if len(labels) > 1:
                    labels = tuple(labels)
                else:
                    labels = labels[0]
                print('FFFF', tuple(features), labels)
                if weight is None:
                    return tuple(features), labels
                else:
                    return tuple(features), labels,weight
            else:
                return tuple(features)
        print('files', self.filenames)
        dataset = tf.data.TFRecordDataset(self.filenames)
        dataset = dataset.map(_parse_function)
        return dataset

    ##
    # @brief
    #       用from_generator的方式生成dataset
    # @param batch_size
    # @param num_epochs
    # @param is_shuffle
    # @param buffer_size
    #
    # @return dataset
    def generate_dataset(self, batch_size=128, num_epochs=1, is_training=True, is_shuffle=False, buffer_size=None):
        """
        生成dataset
        Args:
            batch_size:
            num_epochs:

        Returns:
            dataset
        """
        if self.file_system == 'tfrecord':
            dataset = self._from_tfrecord()
        else:
            if is_training:
                dataset = tf.data.Dataset.from_generator(self.data_generator_training,
                                                         output_shapes=self._get_shapes(is_training),
                                                         output_types=self._get_types(is_training))
            else:
                dataset = tf.data.Dataset.from_generator(self.data_generator,
                                                         output_shapes=self._get_shapes(is_training),
                                                         output_types=self._get_types(is_training))
        if is_shuffle:
            if not buffer_size:
                buffer_size = batch_size * 1000
            dataset = dataset.shuffle(buffer_size=buffer_size)
        dataset = dataset.repeat(num_epochs)
        if self.parser.is_data_padding:
            padded_shapes = self._get_shapes(is_training)
            print('padded_shapes', padded_shapes)
            dataset = dataset.padded_batch(
                batch_size, padded_shapes=padded_shapes)
        else:
            dataset = dataset.batch(batch_size)
        return dataset

    def serialize_example(self, *features):
        """
        Creates a tf.train.Example message ready to be written to a file.
        """
        def _encode_tensor(t):
            return tf.train.Feature(bytes_list=tf.train.BytesList(value=[tf.io.serialize_tensor(t).numpy()]))
        # Create a dictionary mapping the feature name to the tf.train.Example-compatible data type.
        feature = {}
        for f, name in zip(features, self.parser.flat_feature_names):
            feature[name] = _encode_tensor(f)
        # Create a Features message using tf.train.Example
        example_proto = tf.train.Example(
            features=tf.train.Features(feature=feature))
        return example_proto.SerializeToString()

    def to_tfrecords(self, output_path, num=1, num_parallel_calls=8):
        def reduce_func(key, dataset):
            filename = tf.strings.join(
                [output_path, tf.strings.as_string(key)])
            writer = tf.data.experimental.TFRecordWriter(filename)
            writer.write(dataset.map(lambda _, x: x, num_parallel_calls=num_parallel_calls))
            return tf.data.Dataset.from_tensors(filename)

        def tf_serialize_example(*features):
            tf_string = tf.py_function(
                self.serialize_example,
                (features),  # pass these args to the above function.
                tf.string)      # the return type is `tf.string`.
            return tf.reshape(tf_string, ())  # The result is a scalar
        print('types', self._get_types(True, True))
        print('shapes', self._get_shapes(True, True))
        dataset = tf.data.Dataset.from_generator(self.data_generator_tfrecord_flat,
                                                 output_shapes=self._get_shapes(True, True),
                                                 output_types=self._get_types(True, True))
        dataset = dataset.map(tf_serialize_example)
        if num == 1:
            writer = tf.data.experimental.TFRecordWriter(output_path)
            writer.write(dataset)
        elif num > 1:
            dataset = dataset.enumerate()
            dataset = dataset.apply(tf.data.experimental.group_by_window(lambda i, _: i % num, reduce_func, tf.int64.max))
            for _ in enumerate(dataset):
                pass



from __future__ import absolute_import, division, print_function
__all__ = ['DataSchema', 'data_processor_dicts', 'TokenDicts']
from datasets.utils.common_struct import DataSchema
from datasets.utils.data_processor import data_processor_dicts
from datasets.utils.token_dicts import TokenDicts



#!/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020 yinochaos <pspcxl@163.com>. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""raw multi-inputs dataset generator class
"""

from __future__ import absolute_import, division, print_function
import os
import torch
import subprocess
import codecs
import pickle
# import quickle
# import lmdb

from datasets.utils.common_struct import load_hdfs_filelist, load_local_filelist


class PTDataset(torch.utils.data.IterableDataset):
    """从文件流创建的dataset
        利用tf.dataset支持多文件输入（本地和HDFS同时支持），对于大规模数据集十分友好；
        并通过token_dicts和datafields支持配置化的数据处理，灵活支持多输入数据集的处理
        text（local ， hdfs）：当前已经支持
        @TODO : 支持多种文件读取方式
        pickle：finish TODO test
        qiuckle:
        tfrecord：code&test finish
        LMDB（Lightning Memory-Mapped Database(快如闪电的内存映射数据库)）：TODO
        HDF5：TODO
    """
    ##
    # @brief
    #
    # @param file_path dataset文件路径
    # @param token_dicts 分词id词典
    # @param data_field_list dataset的schema数据
    # @param label_field label的schema数据
    # @param is_with_label dataset是否带有label
    # @param file_system
    #               local  : 从本地读取文件
    #               HDFS :  从HDFS读取文件
    # @param file_suffix dataset文件后缀,如果是None则使用所有文件
    # @param weight_fn 计算样本权重的函数 weight_fn(line)
    #
    # @return

    def __init__(self, parser, file_path, file_system='local', file_suffix=None, filenames=None, file_format='text', is_training=True):
        self.is_training = is_training
        self.file_system = file_system
        self.file_format = file_format
        if self.file_system == 'HDFS':
            self.hadoop = os.environ.get('HADOOP')
            if self.hadoop is None:
                raise RuntimeError(
                    'os.environ[HADOOP] is None, should export HADOOP')
        if filenames is None:
            self.filenames = self.load_filelist(file_path, file_suffix, file_system)
            print('file names', self.filenames)
        else:
            self.filenames = filenames
        self.file_suffix = file_suffix
        self.parser = parser
    ##
    # @brief
    #   载入目录所有模型输入文件
    # @param datadir 数据目录
    # @param file_suffix 文件后缀
    # @param file_system
    #
    # @return filenames 所有文件列表

    def load_filelist(self, datadir, file_suffix, file_system):
        filenames = []
        if file_system == 'HDFS':
            filenames = load_hdfs_filelist(datadir, file_suffix, self.hadoop)
        else:
            filenames = load_local_filelist(datadir, file_suffix)
        return filenames

    def read_input_lines(self, filename):
        if self.file_system == 'HDFS':
            cmd = self.hadoop + ' fs -cat ' + filename
            with subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True) as pipe:
                for line in iter(pipe.stdout.readline, b''):
                    yield line.decode('utf8', 'ignore')
        elif self.file_system == 'local':
            with codecs.open(filename, encoding='utf8') as f:
                for line in f:
                    yield line
        elif self.file_system == 'pickle':
            with open(filename, 'rb') as f:
                while True:
                    try:
                        ex = pickle.load(f)
                        if isinstance(ex, list):
                            yield from ex
                        else:
                            yield ex
                    except EOFError:
                        break
        elif self.file_system == 'lmdb':
            pass
            """
            with lmdb.open(filename) as env:
                with env.begin() as txn:
                    cursor = txn.cursor()
                    for _, data in enumerate(cursor.iternext_dup()):
                        if isinstance(data, list):
                            for d in data:
                                yield d
                        else:
                            yield data
            """

    def __iter__(self):
        if self.file_format == 'text':
            for filename in self.filenames:
                for line in self.read_input_lines(filename):
                    label, add_info, features, weight = self.parser.parse(line)
                    if weight is not None and weight == 0:
                        continue
                    if self.is_training:
                        result = ((features, label) if weight is None else (features, label, weight))
                    else:
                        result = ((add_info, features) if label is None else (add_info, features, label))
                    yield result
        elif self.file_system == 'lmdb':
            for filename in self.filenames:
                for line in self.read_input_lines(filename):
                    pass
